# Copyright (c) 2021 Mobvoi Inc. (authors: Binbin Zhang)
#               2023 Horizon Inc. (authors: Xingchen Song)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import Optional
import deepspeed
import json
import logging
import os
import torch
import yaml

import torch.optim as optim
import torch.distributed as dist

from tensorboardX import SummaryWriter
from torch.utils.data import DataLoader
from torch.nn.utils import clip_grad_norm_
from deepspeed.runtime.zero.stage_1_and_2 import (
    estimate_zero2_model_states_mem_needs_all_live)
from deepspeed.runtime.zero.stage3 import (
    estimate_zero3_model_states_mem_needs_all_live)
from deepspeed.utils.zero_to_fp32 import (
    convert_zero_checkpoint_to_fp32_state_dict)
from wenet.dataset.dataset import Dataset
from wenet.utils.checkpoint import save_checkpoint
from wenet.utils.common import StepTimer
from wenet.utils.scheduler import WarmupLR, NoamHoldAnnealing
from wenet.utils.ctc_utils import get_blank_id


def add_model_args(parser):
    parser.add_argument('--config', required=True, help='config file')
    parser.add_argument('--model_dir', required=True, help='save model dir')
    parser.add_argument('--checkpoint', help='checkpoint model')
    parser.add_argument('--tensorboard_dir',
                        default='tensorboard',
                        help='tensorboard log dir')
    parser.add_argument('--override_config',
                        action='append',
                        default=[],
                        help="override yaml config")
    parser.add_argument("--enc_init",
                        default=None,
                        type=str,
                        help="Pre-trained model to initialize encoder")
    parser.add_argument(
        '--enc_init_mods',
        default="encoder.",
        type=lambda s: [str(mod) for mod in s.split(",") if s != ""],
        help="List of encoder modules \
                        to initialize ,separated by a comma")
    parser.add_argument(
        '--freeze_modules',
        default="",
        type=lambda s: [str(mod) for mod in s.split(",") if s != ""],
        help='free module names',
    )
    return parser


def add_trace_args(parser):
    parser.add_argument('--jit',
                        action='store_true',
                        default=False,
                        help='if use jit to trace model while training stage')
    parser.add_argument('--print_model',
                        action='store_true',
                        default=False,
                        help='print model')
    return parser


def add_dataset_args(parser):
    parser.add_argument('--data_type',
                        default='raw',
                        choices=['raw', 'shard'],
                        help='train and cv data type')
    parser.add_argument('--train_data', required=True, help='train data file')
    parser.add_argument('--cv_data', required=True, help='cv data file')
    parser.add_argument('--num_workers',
                        default=0,
                        type=int,
                        help='num of subprocess workers for reading')
    parser.add_argument('--pin_memory',
                        action='store_true',
                        default=False,
                        help='Use pinned memory buffers used for reading')
    parser.add_argument('--prefetch',
                        default=100,
                        type=int,
                        help='prefetch number')
    return parser


def add_ddp_args(parser):
    parser.add_argument('--ddp.dist_backend',
                        dest='dist_backend',
                        default='nccl',
                        choices=['nccl', 'gloo'],
                        help='distributed backend')
    parser.add_argument('--use_amp',
                        action='store_true',
                        default=False,
                        help='Use automatic mixed precision training')
    parser.add_argument('--fp16_grad_sync',
                        action='store_true',
                        default=False,
                        help='Use fp16 gradient sync for ddp')
    return parser


def add_deepspeed_args(parser):
    parser.add_argument('--timeout',
                        default=30,
                        type=int,
                        help='timeout (in seconds) of wenet_join. ' +
                        '30s for aishell & 300s for wenetspeech')
    parser.add_argument('--local_rank',
                        type=int,
                        default=-1,
                        help='local rank passed from distributed launcher')
    parser.add_argument('--deepspeed.save_states',
                        dest='save_states',
                        default='model_only',
                        choices=['model_only', 'model+optimizer'],
                        help='save model/optimizer states')
    # DeepSpeed automaticly add '--deepspeed' and '--deepspeed_config' to parser
    parser = deepspeed.add_config_arguments(parser)
    return parser


def init_distributed(args):
    world_size = int(os.environ.get('WORLD_SIZE', 1))
    local_rank = int(os.environ.get('LOCAL_RANK', 0))
    rank = int(os.environ.get('RANK', 0))
    logging.info('training on multiple gpus, this gpu {}'.format(local_rank) +
                 ', rank {}, world_size {}'.format(rank, world_size))
    if args.train_engine == "torch_ddp":
        torch.cuda.set_device(local_rank)
        dist.init_process_group(args.dist_backend)
    elif args.train_engine == "deepspeed":
        deepspeed.init_distributed(dist_backend=args.dist_backend)
    else:
        logging.error("not supported engine: {}".format(args.train_engine))
    return world_size, local_rank, rank


def check_modify_and_save_config(args, configs, symbol_table):
    if args.train_engine == "torch_ddp":
        if args.use_amp:
            configs["dtype"] = "fp16"
        else:
            configs["dtype"] = "fp32"
    elif args.train_engine == "deepspeed":
        # NOTE(xcsong): DeepSpeed does not support uneven data. When using custom
        #   dataset, we need to manually ensure that the data is evenly distributed
        #   across all processe. we impl `train_utils.py::wenet_join` for this func
        #   ref: https://github.com/microsoft/DeepSpeed/issues/2223
        #
        # NOTE(xsong):  We also need to keep:
        #       1. `train_micro_batch_size_per_gpu == 1`
        #       2. `accum_grad (in train_confomrer.yaml)
        #               == gradient_accumulation_steps (in ds_config.json)`
        #       3. `grad_clip (in train_confomrer.yaml)
        #               == gradient_clipping (in ds_config.json)`
        #   The reason for such consistence checking lies in that deepspeed's native
        #   dataloader uses PyTorch's torch.utils.py.data.DistributedSampler which does
        #   not support IterableDataset, IterableDataset is extremly useful in large
        #   scale training because it lets you stream the data without having to
        #   download the complete dataset.
        #       ref: https://github.com/microsoft/DeepSpeed/issues/1371
        #           https://github.com/microsoft/DeepSpeed/issues/285
        #   To make deepspeed training compatible with IterableDataset, we have to
        #   use custom dataloader instead of deepspeed's native loader and thus we
        #   should configure batchsize in train_confomrer.yaml instead of
        #   ds_config.json. On the contrary, gradient accumulation / clipping should be
        #   configured in ds_config.json since they will be handled by ds automatically.
        #       ref: https://github.com/microsoft/DeepSpeed/issues/62
        with open(args.deepspeed_config, 'r') as fin:
            ds_configs = json.load(fin)
        if "fp16" in ds_configs and ds_configs["fp16"]["enabled"]:
            configs["dtype"] = "fp16"
        elif "bf16" in ds_configs and ds_configs["bf16"]["enabled"]:
            configs["dtype"] = "bf16"
        else:
            configs["dtype"] = "fp32"
        assert ds_configs["train_micro_batch_size_per_gpu"] == 1
        assert ds_configs["gradient_accumulation_steps"] == configs[
            'accum_grad']
        assert ds_configs["gradient_clipping"] == configs['grad_clip']
        assert ds_configs["steps_per_print"] == configs['log_interval']

    if 'input_dim' not in configs:
        if 'fbank_conf' in configs['dataset_conf']:
            input_dim = configs['dataset_conf']['fbank_conf']['num_mel_bins']
        elif 'log_mel_spectrogram_conf' in configs['dataset_conf']:
            input_dim = configs['dataset_conf']['log_mel_spectrogram_conf'][
                'num_mel_bins']
        else:
            input_dim = configs['dataset_conf']['mfcc_conf']['num_mel_bins']
    else:
        input_dim = configs['input_dim']

    configs, _ = get_blank_id(configs, symbol_table)

    configs['input_dim'] = input_dim
    configs['output_dim'] = configs['vocab_size']

    configs['train_engine'] = args.train_engine
    configs['use_amp'] = args.use_amp
    configs['model_dir'] = args.model_dir
    configs['save_states'] = args.save_states

    # Save configs to model_dir/train.yaml for inference and export
    if int(os.environ.get('RANK', 0)) == 0:
        saved_config_path = os.path.join(args.model_dir, 'train.yaml')
        with open(saved_config_path, 'w') as fout:
            data = yaml.dump(configs)
            fout.write(data)

    if configs["model_conf"].get("apply_non_blank_embedding", False):
        logging.warn('Had better load a well trained model'
                     'if apply_non_blank_embedding is true !!!')

    return configs


def init_dataset_and_dataloader(args, configs, tokenizer, seed=777):
    generator = torch.Generator()
    generator.manual_seed(seed)

    # if save_interval in configs, steps mode else epoch mode
    if "save_interval" in configs:
        configs['dataset_conf']['cycle'] = configs.get('max_epoch', 100)
    train_conf = configs['dataset_conf']
    cv_conf = copy.deepcopy(train_conf)
    cv_conf['cycle'] = 1
    cv_conf['speed_perturb'] = False
    cv_conf['spec_aug'] = False
    cv_conf['spec_sub'] = False
    cv_conf['spec_trim'] = False
    cv_conf['shuffle'] = False

    configs['vocab_size'] = tokenizer.vocab_size()
    train_dataset = Dataset(args.data_type, args.train_data, tokenizer,
                            train_conf, True)
    cv_dataset = Dataset(args.data_type,
                         args.cv_data,
                         tokenizer,
                         cv_conf,
                         partition=False)

    # NOTE(xcsong): Why we prefer persistent_workers=True ?
    #   https://discuss.pytorch.org/t/what-are-the-dis-advantages-of-persistent-workers/102110
    train_data_loader = DataLoader(train_dataset,
                                   batch_size=None,
                                   pin_memory=args.pin_memory,
                                   num_workers=args.num_workers,
                                   persistent_workers=True,
                                   generator=generator,
                                   prefetch_factor=args.prefetch)
    cv_data_loader = DataLoader(cv_dataset,
                                batch_size=None,
                                pin_memory=args.pin_memory,
                                num_workers=args.num_workers,
                                persistent_workers=True,
                                generator=generator,
                                prefetch_factor=args.prefetch)
    return train_dataset, cv_dataset, train_data_loader, cv_data_loader


def wrap_cuda_model(args, model):
    local_world_size = int(os.environ.get('LOCAL_WORLD_SIZE', 1))
    world_size = int(os.environ.get('WORLD_SIZE', 1))
    if hasattr(model, 'encoder'):
        grad_ckpt = getattr(model.encoder, 'gradient_checkpointing', False)
    else:
        grad_ckpt = False
    # TODO(xcsong): could one GPU use ddp? and int(os.environ.get('WORLD_SIZE', 1)) > 1
    if args.train_engine == "torch_ddp":  # native pytorch ddp
        assert (torch.cuda.is_available())
        model.cuda()
        model = torch.nn.parallel.DistributedDataParallel(
            model, find_unused_parameters=not grad_ckpt)
        device = torch.device("cuda")
        if args.fp16_grad_sync:
            from torch.distributed.algorithms.ddp_comm_hooks import (
                default as comm_hooks, )
            model.register_comm_hook(state=None,
                                     hook=comm_hooks.fp16_compress_hook)
    elif args.train_engine == "deepspeed":  # deepspeed
        # NOTE(xcsong): look in detail how the memory estimator API works:
        #   https://deepspeed.readthedocs.io/en/latest/memory.html#discussion
        if int(os.environ.get('RANK', 0)) == 0:
            logging.info("Estimating model states memory needs (zero2)...")
            estimate_zero2_model_states_mem_needs_all_live(
                model,
                num_gpus_per_node=local_world_size,
                num_nodes=world_size // local_world_size)
            logging.info("Estimating model states memory needs (zero3)...")
            estimate_zero3_model_states_mem_needs_all_live(
                model,
                num_gpus_per_node=local_world_size,
                num_nodes=world_size // local_world_size)
        device = None  # Init device later
        pass  # Init DeepSpeed later
    else:
        logging.error("not supported engine: {}".format(args.train_engine))

    return model, device


def init_optimizer_and_scheduler(args, configs, model):
    if configs['optim'] == 'adam':
        optimizer = optim.Adam(model.parameters(), **configs['optim_conf'])
    elif configs['optim'] == 'adamw':
        optimizer = optim.AdamW(model.parameters(), **configs['optim_conf'])
    else:
        raise ValueError("unknown optimizer: " + configs['optim'])

    scheduler_type = None
    if configs['scheduler'] == 'warmuplr':
        scheduler_type = WarmupLR
        scheduler = WarmupLR(optimizer, **configs['scheduler_conf'])
    elif configs['scheduler'] == 'NoamHoldAnnealing':
        scheduler_type = NoamHoldAnnealing
        scheduler = NoamHoldAnnealing(optimizer, **configs['scheduler_conf'])
    else:
        raise ValueError("unknown scheduler: " + configs['scheduler'])

    # NOTE(xcsong): Custom optimizer might yield poor performance when
    #   zero-offload is enabled, if you do want to offload optimizer to CPU,
    #   please set optimizer in ds_config.json, see:
    #   (https://www.deepspeed.ai/docs/config-json/#optimizer-parameters)
    if args.train_engine == "deepspeed":
        with open(args.deepspeed_config, 'r') as fin:
            ds_configs = json.load(fin)
        if "optimizer" in ds_configs:
            # NOTE(xcsong): Disable custom optimizer if it is set in ds_config,
            # extremely useful when enable cpu_offload, DeepspeedCpuAdam
            # could be 4~5x faster than torch native adam
            optimizer = None
            if "scheduler" in ds_configs:
                scheduler = None
            else:

                def scheduler(opt):
                    return scheduler_type(opt, **configs['scheduler_conf'])

        model, optimizer, _, scheduler = deepspeed.initialize(
            args=args,
            model=model,
            optimizer=optimizer,
            lr_scheduler=scheduler,
            model_parameters=model.parameters())

    step = configs["init_infos"].get("step", -1)
    scheduler.set_step(step)
    return model, optimizer, scheduler


def trace_and_print_model(args, model):
    # !!!IMPORTANT!!!
    # Try to export the model by script, if fails, we should refine
    # the code to satisfy the script export requirements
    if int(os.environ.get('RANK', 0)) == 0:
        if args.jit:
            script_model = torch.jit.script(model)
            script_model.save(os.path.join(args.model_dir, 'init.zip'))
        if args.print_model:
            print(model)
            num_params = sum(p.numel() for p in model.parameters())
            print('the number of model params: {:,d}'.format(num_params))


def init_summarywriter(args):
    writer = None
    if int(os.environ.get('RANK', 0)) == 0:
        os.makedirs(args.model_dir, exist_ok=True)
        exp_id = os.path.basename(args.model_dir)
        writer = SummaryWriter(os.path.join(args.tensorboard_dir, exp_id))
    return writer


def save_model(model, info_dict):
    rank = int(os.environ.get('RANK', 0))
    tag = info_dict["tag"]
    model_dir = info_dict["model_dir"]
    # save ckpt
    if info_dict["train_engine"] == "deepspeed":
        # NOTE(xcsong): All ranks should call this API, but only rank 0
        #   save the general model params. see:
        #   https://github.com/microsoft/DeepSpeed/issues/2993
        with torch.no_grad():
            model.save_checkpoint(save_dir=model_dir,
                                  tag=tag,
                                  client_state=info_dict)
            if info_dict["save_states"] == "model_only" and rank == 0:
                convert_zero_checkpoint_to_fp32_state_dict(model_dir,
                                                           "{}/{}.pt".format(
                                                               model_dir, tag),
                                                           tag=tag)
                os.system("rm -rf {}/{}".format(model_dir, tag))
    elif rank == 0:
        # NOTE(xcsong): For torch_ddp, only rank-0 should call this.
        save_model_path = os.path.join(model_dir, '{}.pt'.format(tag))
        save_checkpoint(model, save_model_path, info_dict)
    # save yaml
    if rank == 0:
        with open("{}/{}.yaml".format(model_dir, tag), 'w') as fout:
            data = yaml.dump(info_dict)
            fout.write(data)


def wenet_join(group_join, info_dict):
    world_size = int(os.environ.get('WORLD_SIZE', 1))
    local_rank = int(os.environ.get('LOCAL_RANK', 0))
    rank = int(os.environ.get('RANK', 0))
    train_engine = info_dict.get('train_engine', "torch_ddp")

    if info_dict["batch_idx"] == 0 or train_engine == "torch_ddp":
        # NOTE(xcsong): skip first batch because its processing time includes
        #   dataloader initialization time, which may exceed 30 seconds
        return False

    try:
        # NOTE(xcsong): Why we need a new group?
        #   Because Deepspeed has its own group where all the relevant communication
        #   operations are executed. If we add a communication operation that is not
        #   managed by Deepspeed in this group, it's highly likely to cause
        #   communication chaos, resulting in hard-to-troubleshoot hangs.
        dist.monitored_barrier(group=group_join,
                               timeout=group_join.options._timeout)
    except RuntimeError as e:
        logging.info("Detected uneven workload distribution: {}\n".format(e) +
                     "Break current worker to manually join all workers, " +
                     "world_size {}, current rank {}, current local_rank {}\n".
                     format(world_size, rank, local_rank))
        return True

    return False


def batch_forward(model, batch, scaler, info_dict):
    train_engine = info_dict.get('train_engine', "torch_ddp")
    device = int(os.environ.get('LOCAL_RANK', 0))
    accum_grad = info_dict.get('accum_grad', 1)

    dtype = info_dict.get("dtype", "fp32")
    if dtype == "fp16":
        dtype = torch.float16
    elif dtype == "bf16":
        dtype = torch.bfloat16
    else:  # fp32
        dtype = None

    if train_engine == "deepspeed":
        # deepspeed
        with torch.cuda.amp.autocast(enabled=dtype is not None,
                                     dtype=dtype,
                                     cache_enabled=False):
            loss_dict = model(batch, device)
    else:
        # torch_ddp
        # autocast context
        # The more details about amp can be found in
        # https://pytorch.org/docs/stable/notes/amp_examples.html
        with torch.cuda.amp.autocast(scaler is not None):
            loss_dict = model(batch, device)
    info_dict['loss_dict'] = loss_dict

    return info_dict


def batch_backward(model, scaler, info_dict):
    train_engine = info_dict.get("train_engine", "torch_ddp")
    accum_grad = info_dict.get('accum_grad', 1)
    use_amp = info_dict.get('use_amp', False)
    if use_amp:
        assert scaler is not None
    loss = info_dict['loss_dict']['loss']

    if train_engine == "deepspeed":
        # NOTE(xcsong): `model.backward(loss)` is equivalent to
        #               `scale_loss_wrt_accum_grad + loss.backward()`
        #   ref: https://www.deepspeed.ai/tutorials/megatron/#using-the-training-api
        scaled_loss = model.backward(loss)
    elif train_engine == "torch_ddp":
        scaled_loss = loss / accum_grad
        if use_amp:
            scaler.scale(scaled_loss).backward()
        else:
            scaled_loss.backward()
    info_dict['loss_dict']['loss'] = scaled_loss
    for loss_name, loss_value in info_dict['loss_dict'].items():
        if loss_value is not None:
            info_dict['loss_dict'][loss_name] = loss_value.item()

    return info_dict


def update_parameter_and_lr(model, optimizer, scheduler, scaler, info_dict):
    rank = int(os.environ.get('RANK', 0))
    train_engine = info_dict.get("train_engine", "torch_ddp")
    accum_grad = info_dict.get('accum_grad', 1)
    use_amp = info_dict.get('use_amp', False)
    clip = info_dict.get('grad_clip', 50.0)
    batch_idx = info_dict["batch_idx"]
    if use_amp:
        assert scaler is not None

    grad_norm = 0.0
    if train_engine == "deepspeed":
        # NOTE(xcsong): The step() function in DeepSpeed engine updates the
        #   model parameters as well as the learning rate.
        #   Zeroing the gradients is handled automatically by
        #   DeepSpeed after the weights have been updated using a mini-batch.
        #   DeepSpeed also performs gradient averaging automatically at the
        #   gradient accumulation boundaries and addresses clip_grad_norm internally.
        #   `ds_model.step() =  clip_grad_norm_() + optimizer.step()
        #                       + optimizer.zero_grad() + scheduler.step()`
        #   ref: https://www.deepspeed.ai/tutorials/megatron/#using-the-training-api
        info_dict["is_gradient_accumulation_boundary"] = \
            model.is_gradient_accumulation_boundary()
        model.step()
        grad_norm = model.get_global_grad_norm()
    elif (batch_idx + 1) % accum_grad == 0:
        # Use mixed precision training
        if use_amp:
            scaler.unscale_(optimizer)
            grad_norm = clip_grad_norm_(model.parameters(), clip)
            # Must invoke scaler.update() if unscale_() is used in
            # the iteration to avoid the following error:
            #   RuntimeError: unscale_() has already been called
            #   on this optimizer since the last update().
            # We don't check grad here since that if the gradient
            # has inf/nan values, scaler.step will skip
            # optimizer.step().
            scaler.step(optimizer)
            scaler.update()
        else:
            grad_norm = clip_grad_norm_(model.parameters(), clip)
            if torch.isfinite(grad_norm):
                optimizer.step()
        optimizer.zero_grad()
        scheduler.step()
        grad_norm = grad_norm.item()

    info_dict["lr"] = optimizer.param_groups[0]['lr']
    info_dict["grad_norm"] = grad_norm

    return info_dict


def log_per_step(writer, info_dict, timer: Optional[StepTimer] = None):
    tag = info_dict["tag"]
    step = info_dict["step"]
    batch_idx = info_dict["batch_idx"]
    loss_dict = info_dict['loss_dict']
    epoch = info_dict.get('epoch', 0)
    train_engine = info_dict.get("train_engine", "torch_ddp")
    accum_grad = info_dict.get('accum_grad', 1) if tag != "CV" else 1
    log_interval = info_dict.get('log_interval', 10)
    lr = info_dict.get("lr", 0.0)
    is_gradient_accumulation_boundary = info_dict.get(
        "is_gradient_accumulation_boundary", False)

    rank = int(os.environ.get('RANK', 0))

    if tag == "TRAIN" and rank == 0 and writer is not None:
        if (train_engine == "deepspeed" and is_gradient_accumulation_boundary) or \
           (train_engine == "torch_ddp" and (batch_idx + 1) % accum_grad == 0):
            writer.add_scalar('train/train_loss',
                              loss_dict['loss'] * accum_grad, step + 1)
            writer.add_scalar('train/grad_norm', info_dict['grad_norm'],
                              step + 1)
            for name, value in loss_dict.items():
                if name != 'loss' and value is not None:
                    writer.add_scalar('train/{}'.format(name), value, step + 1)
    elif "step_" in tag and rank == 0 and writer is not None:
        writer.add_scalar('global_step/lr', lr, step + 1)
        for name, value in loss_dict.items():
            writer.add_scalar('global_step/{}'.format(name), value, step + 1)

    if (batch_idx + 1) % log_interval == 0:
        log_str = '{} | '.format(tag)
        if timer is not None:
            timer_step = step
            if info_dict.get("cv_step", None) is not None:
                timer_step = info_dict['cv_step']
            steps_per_second = timer.steps_per_second(timer_step)
            log_str += 'steps/sec {:.1f}| '.format(steps_per_second)
        log_str += 'Batch {}/{} loss {:.6f} '.format(
            epoch, batch_idx + 1, loss_dict['loss'] * accum_grad)
        for name, value in loss_dict.items():
            if name != 'loss' and value is not None:
                log_str += '{} {:.6f} '.format(name, value)
        if tag == "TRAIN":
            log_str += 'lr {:.8f} grad_norm {:.6f} rank {}'.format(
                lr, info_dict['grad_norm'], rank)
        logging.debug(log_str)


def log_per_epoch(writer, info_dict):
    epoch = info_dict["epoch"]
    loss_dict = info_dict["loss_dict"]
    if int(os.environ.get('RANK', 0)) == 0:
        writer.add_scalar('epoch/lr', info_dict["lr"], epoch)
        for name, value in loss_dict.items():
            writer.add_scalar('epoch/{}'.format(name), value, epoch)


def freeze_modules(model, args):
    for name, param in model.named_parameters():
        for module_name in args.freeze_modules:
            if module_name in name:
                param.requires_grad = False
                logging.debug("{} module is freezed".format(name))



from typing import List

import k2
import torch
from torch import nn

from icefall.mmi_graph_compiler import MmiTrainingGraphCompiler


def _compute_mmi_loss_exact_optimized(
    dense_fsa_vec: k2.DenseFsaVec,
    texts: List[str],
    graph_compiler: MmiTrainingGraphCompiler,
    den_scale: float = 1.0,
    beam_size: float = 8.0,
) -> torch.Tensor:
    """
    The function name contains `exact`, which means it uses a version of
    intersection without pruning.

    `optimized` in the function name means this function is optimized
    in that it calls k2.intersect_dense only once

    Note:
      It is faster at the cost of using more memory.

    Args:
      dense_fsa_vec:
        It contains the neural network output.
      texts:
        The transcript. Each element consists of space(s) separated words.
      graph_compiler:
        Used to build num_graphs and den_graphs
      den_scale:
        The scale applied to the denominator tot_scores.
    Returns:
      Return a scalar loss. It is the sum over utterances in a batch,
      without normalization.
    """
    num_graphs, den_graphs = graph_compiler.compile(texts, replicate_den=False)

    device = num_graphs.device

    num_fsas = num_graphs.shape[0]
    assert dense_fsa_vec.dim0() == num_fsas

    assert den_graphs.shape[0] == 1

    # The motivation to concatenate num_graphs and den_graphs
    # is to reduce the number of calls to k2.intersect_dense.
    num_den_graphs = k2.cat([num_graphs, den_graphs])

    # NOTE: The a_to_b_map in k2.intersect_dense must be sorted
    # so the following reorders num_den_graphs.
    #
    # The following code computes a_to_b_map

    # [0, 1, 2, ... ]
    num_graphs_indexes = torch.arange(num_fsas, dtype=torch.int32)

    # [num_fsas, num_fsas, num_fsas, ... ]
    den_graphs_indexes = torch.tensor([num_fsas] * num_fsas, dtype=torch.int32)

    # [0, num_fsas, 1, num_fsas, 2, num_fsas, ... ]
    num_den_graphs_indexes = (
        torch.stack([num_graphs_indexes, den_graphs_indexes]).t().reshape(-1).to(device)
    )

    num_den_reordered_graphs = k2.index(num_den_graphs, num_den_graphs_indexes)

    # [[0, 1, 2, ...]]
    a_to_b_map = torch.arange(num_fsas, dtype=torch.int32).reshape(1, -1)

    # [[0, 1, 2, ...]] -> [0, 0, 1, 1, 2, 2, ... ]
    a_to_b_map = a_to_b_map.repeat(2, 1).t().reshape(-1).to(device)

    num_den_lats = k2.intersect_dense(
        num_den_reordered_graphs,
        dense_fsa_vec,
        output_beam=beam_size,
        a_to_b_map=a_to_b_map,
    )

    num_den_tot_scores = num_den_lats.get_tot_scores(
        log_semiring=True, use_double_scores=True
    )

    num_tot_scores = num_den_tot_scores[::2]
    den_tot_scores = num_den_tot_scores[1::2]

    tot_scores = num_tot_scores - den_scale * den_tot_scores
    loss = -1 * tot_scores.sum()
    return loss


def _compute_mmi_loss_exact_non_optimized(
    dense_fsa_vec: k2.DenseFsaVec,
    texts: List[str],
    graph_compiler: MmiTrainingGraphCompiler,
    den_scale: float = 1.0,
    beam_size: float = 8.0,
) -> torch.Tensor:
    """
    See :func:`_compute_mmi_loss_exact_optimized` for the meaning
    of the arguments.

    It's more readable, though it invokes k2.intersect_dense twice.

    Note:
      It uses less memory at the cost of speed. It is slower.
    """
    num_graphs, den_graphs = graph_compiler.compile(texts, replicate_den=True)

    # TODO: pass output_beam as function argument
    num_lats = k2.intersect_dense(
        num_graphs, dense_fsa_vec, output_beam=beam_size, max_arcs=2147483600
    )
    den_lats = k2.intersect_dense(
        den_graphs, dense_fsa_vec, output_beam=beam_size, max_arcs=2147483600
    )

    num_tot_scores = num_lats.get_tot_scores(log_semiring=True, use_double_scores=True)

    den_tot_scores = den_lats.get_tot_scores(log_semiring=True, use_double_scores=True)

    tot_scores = num_tot_scores - den_scale * den_tot_scores

    loss = -1 * tot_scores.sum()
    return loss


def _compute_mmi_loss_pruned(
    dense_fsa_vec: k2.DenseFsaVec,
    texts: List[str],
    graph_compiler: MmiTrainingGraphCompiler,
    den_scale: float = 1.0,
    beam_size: float = 8.0,
) -> torch.Tensor:
    """
    See :func:`_compute_mmi_loss_exact_optimized` for the meaning
    of the arguments.

    `pruned` means it uses k2.intersect_dense_pruned

    Note:
      It uses the least amount of memory, but the loss is not exact due
      to pruning.
    """
    num_graphs, den_graphs = graph_compiler.compile(texts, replicate_den=False)

    num_lats = k2.intersect_dense(num_graphs, dense_fsa_vec, output_beam=8.0)

    # the values for search_beam/output_beam/min_active_states/max_active_states
    # are not tuned. You may want to tune them.
    den_lats = k2.intersect_dense_pruned(
        den_graphs,
        dense_fsa_vec,
        search_beam=20.0,
        output_beam=beam_size,
        min_active_states=30,
        max_active_states=10000,
    )

    num_tot_scores = num_lats.get_tot_scores(log_semiring=True, use_double_scores=True)

    den_tot_scores = den_lats.get_tot_scores(log_semiring=True, use_double_scores=True)

    tot_scores = num_tot_scores - den_scale * den_tot_scores

    loss = -1 * tot_scores.sum()
    return loss


class LFMMILoss(nn.Module):
    """
    Computes Lattice-Free Maximum Mutual Information (LFMMI) loss.

    TODO: more detailed description
    """

    def __init__(
        self,
        graph_compiler: MmiTrainingGraphCompiler,
        use_pruned_intersect: bool = False,
        den_scale: float = 1.0,
        beam_size: float = 8.0,
    ):
        super().__init__()
        self.graph_compiler = graph_compiler
        self.den_scale = den_scale
        self.use_pruned_intersect = use_pruned_intersect
        self.beam_size = beam_size

    def forward(
        self,
        dense_fsa_vec: k2.DenseFsaVec,
        texts: List[str],
    ) -> torch.Tensor:
        """
        Args:
          dense_fsa_vec:
            It contains the neural network output.
          texts:
            A list of strings. Each string contains space(s) separated words.
        Returns:
          Return a scalar loss. It is the sum over utterances in a batch,
          without normalization.
        """
        if self.use_pruned_intersect:
            func = _compute_mmi_loss_pruned
        else:
            func = _compute_mmi_loss_exact_non_optimized
            #  func = _compute_mmi_loss_exact_optimized

        return func(
            dense_fsa_vec=dense_fsa_vec,
            texts=texts,
            graph_compiler=self.graph_compiler,
            den_scale=self.den_scale,
            beam_size=self.beam_size,
        )



import tqdm

from gxl_ai_utils.utils import utils_file

def little_func(input_dict, res_dict, output_dir):
    temp_res_dict = {}
    for k, wav_path in tqdm.tqdm(input_dict.items(), total=len(input_dict), desc=f"移动音频中"):
        output_wav_path = f"{output_dir}/{k}.npy"
        utils_file.copy_file(wav_path, output_wav_path, use_shell=True)
        temp_res_dict[k] = output_wav_path
    res_dict.update(temp_res_dict)



def get_wav_file():
    intput_dir = "/home/work_nfs8/xlgeng/new_workspace/gxl_ai_utils/eggs/cats_and_dogs/prepare_data_for_codec_asr/output2"
    partition = ['train', 'dev', 'test']
    for p in partition:
        output_wav_dir = f"{intput_dir}/{p}/wav"
        input_data_path = f"{intput_dir}/{p}/wav_old.scp"
        wav_path_dict = utils_file.load_dict_from_scp(input_data_path)
        utils_file.write_dict_to_scp(wav_path_dict, f"{intput_dir}/{p}/wav_old.scp")
        new_wav_dict = {}
        dict_list = utils_file.do_split_dict(wav_path_dict, 45)
        runner = utils_file.GxlDynamicThreadPool()
        for dict_i in dict_list:
            runner.add_task(little_func, [dict_i, new_wav_dict, output_wav_dir])
        runner.start()
        utils_file.write_dict_to_scp(new_wav_dict, f"{intput_dir}/{p}/wav.scp")
        utils_file.do_convert_wav_text_scp_to_jsonl(wav_scp_file_path=f"{intput_dir}/{p}/wav.scp",
                                                    text_scp_file_path=f"{intput_dir}/{p}/text",
                                                    target_jsonl_file_path=f"{intput_dir}/{p}/data.list")

if __name__ == '__main__':
    get_wav_file()


# Copyright (c) 2021 Mobvoi Inc (Chao Yang)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import random
import math

import torchaudio
import torch


def db2amp(db):
    return pow(10, db / 20)


def amp2db(amp):
    return 20 * math.log10(amp)


def make_poly_distortion(conf):
    """Generate a db-domain ploynomial distortion function

        f(x) = a * x^m * (1-x)^n + x

    Args:
        conf: a dict {'a': #int, 'm': #int, 'n': #int}

    Returns:
        The ploynomial function, which could be applied on
        a float amplitude value
    """
    a = conf['a']
    m = conf['m']
    n = conf['n']

    def poly_distortion(x):
        abs_x = abs(x)
        if abs_x < 0.000001:
            x = x
        else:
            db_norm = amp2db(abs_x) / 100 + 1
            if db_norm < 0:
                db_norm = 0
            db_norm = a * pow(db_norm, m) * pow((1 - db_norm), n) + db_norm
            if db_norm > 1:
                db_norm = 1
            db = (db_norm - 1) * 100
            amp = db2amp(db)
            if amp >= 0.9997:
                amp = 0.9997
            if x > 0:
                x = amp
            else:
                x = -amp
        return x

    return poly_distortion


def make_quad_distortion():
    return make_poly_distortion({'a': 1, 'm': 1, 'n': 1})


# the amplitude are set to max for all non-zero point
def make_max_distortion(conf):
    """Generate a max distortion function

    Args:
        conf: a dict {'max_db': float }
            'max_db': the maxium value.

    Returns:
        The max function, which could be applied on
        a float amplitude value
    """
    max_db = conf['max_db']
    if max_db:
        max_amp = db2amp(max_db)  # < 0.997
    else:
        max_amp = 0.997

    def max_distortion(x):
        if x > 0:
            x = max_amp
        elif x < 0:
            x = -max_amp
        else:
            x = 0.0
        return x

    return max_distortion


def make_amp_mask(db_mask=None):
    """Get a amplitude domain mask from db domain mask

    Args:
        db_mask: Optional. A list of tuple. if None, using default value.

    Returns:
        A list of tuple. The amplitude domain mask
    """
    if db_mask is None:
        db_mask = [(-110, -95), (-90, -80), (-65, -60), (-50, -30), (-15, 0)]
    amp_mask = [(db2amp(db[0]), db2amp(db[1])) for db in db_mask]
    return amp_mask


default_mask = make_amp_mask()


def generate_amp_mask(mask_num):
    """Generate amplitude domain mask randomly in [-100db, 0db]

    Args:
        mask_num: the slot number of the mask

    Returns:
        A list of tuple. each tuple defines a slot.
        e.g. [(-100, -80), (-65, -60), (-50, -30), (-15, 0)]
        for #mask_num = 4
    """
    a = [0] * 2 * mask_num
    a[0] = 0
    m = []
    for i in range(1, 2 * mask_num):
        a[i] = a[i - 1] + random.uniform(0.5, 1)
    max_val = a[2 * mask_num - 1]
    for i in range(0, mask_num):
        l = ((a[2 * i] - max_val) / max_val) * 100
        r = ((a[2 * i + 1] - max_val) / max_val) * 100
        m.append((l, r))
    return make_amp_mask(m)


def make_fence_distortion(conf):
    """Generate a fence distortion function

    In this fence-like shape function, the values in mask slots are
    set to maxium, while the values not in mask slots are set to 0.
    Use seperated masks for Positive and negetive amplitude.

    Args:
        conf: a dict {'mask_number': int,'max_db': float }
            'mask_number': the slot number in mask.
            'max_db': the maxium value.

    Returns:
        The fence function, which could be applied on
        a float amplitude value
    """
    mask_number = conf['mask_number']
    max_db = conf['max_db']
    max_amp = db2amp(max_db)  # 0.997
    if mask_number <= 0:
        positive_mask = default_mask
        negative_mask = make_amp_mask([(-50, 0)])
    else:
        positive_mask = generate_amp_mask(mask_number)
        negative_mask = generate_amp_mask(mask_number)

    def fence_distortion(x):
        is_in_mask = False
        if x > 0:
            for mask in positive_mask:
                if x >= mask[0] and x <= mask[1]:
                    is_in_mask = True
                    return max_amp
            if not is_in_mask:
                return 0.0
        elif x < 0:
            abs_x = abs(x)
            for mask in negative_mask:
                if abs_x >= mask[0] and abs_x <= mask[1]:
                    is_in_mask = True
                    return max_amp
            if not is_in_mask:
                return 0.0
        return x

    return fence_distortion


#
def make_jag_distortion(conf):
    """Generate a jag distortion function

    In this jag-like shape function, the values in mask slots are
    not changed, while the values not in mask slots are set to 0.
    Use seperated masks for Positive and negetive amplitude.

    Args:
        conf: a dict {'mask_number': #int}
            'mask_number': the slot number in mask.

    Returns:
        The jag function,which could be applied on
        a float amplitude value
    """
    mask_number = conf['mask_number']
    if mask_number <= 0:
        positive_mask = default_mask
        negative_mask = make_amp_mask([(-50, 0)])
    else:
        positive_mask = generate_amp_mask(mask_number)
        negative_mask = generate_amp_mask(mask_number)

    def jag_distortion(x):
        is_in_mask = False
        if x > 0:
            for mask in positive_mask:
                if x >= mask[0] and x <= mask[1]:
                    is_in_mask = True
                    return x
            if not is_in_mask:
                return 0.0
        elif x < 0:
            abs_x = abs(x)
            for mask in negative_mask:
                if abs_x >= mask[0] and abs_x <= mask[1]:
                    is_in_mask = True
                    return x
            if not is_in_mask:
                return 0.0
        return x

    return jag_distortion


# gaining 20db means amp = amp * 10
# gaining -20db means amp = amp / 10
def make_gain_db(conf):
    """Generate a db domain gain function

    Args:
        conf: a dict {'db': #float}
            'db': the gaining value

    Returns:
        The db gain function, which could be applied on
        a float amplitude value
    """
    db = conf['db']

    def gain_db(x):
        return min(0.997, x * pow(10, db / 20))

    return gain_db


def distort(x, func, rate=0.8):
    """Distort a waveform in sample point level

    Args:
        x: the origin wavefrom
        func: the distort function
        rate: sample point-level distort probability

    Returns:
        the distorted waveform
    """
    for i in range(0, x.shape[1]):
        a = random.uniform(0, 1)
        if a < rate:
            x[0][i] = func(float(x[0][i]))
    return x


def distort_chain(x, funcs, rate=0.8):
    for i in range(0, x.shape[1]):
        a = random.uniform(0, 1)
        if a < rate:
            for func in funcs:
                x[0][i] = func(float(x[0][i]))
    return x


# x is numpy
def distort_wav_conf(x, distort_type, distort_conf, rate=0.1):
    if distort_type == 'gain_db':
        gain_db = make_gain_db(distort_conf)
        x = distort(x, gain_db)
    elif distort_type == 'max_distortion':
        max_distortion = make_max_distortion(distort_conf)
        x = distort(x, max_distortion, rate=rate)
    elif distort_type == 'fence_distortion':
        fence_distortion = make_fence_distortion(distort_conf)
        x = distort(x, fence_distortion, rate=rate)
    elif distort_type == 'jag_distortion':
        jag_distortion = make_jag_distortion(distort_conf)
        x = distort(x, jag_distortion, rate=rate)
    elif distort_type == 'poly_distortion':
        poly_distortion = make_poly_distortion(distort_conf)
        x = distort(x, poly_distortion, rate=rate)
    elif distort_type == 'quad_distortion':
        quad_distortion = make_quad_distortion()
        x = distort(x, quad_distortion, rate=rate)
    elif distort_type == 'none_distortion':
        pass
    else:
        print('unsupport type')
    return x


def distort_wav_conf_and_save(distort_type, distort_conf, rate, wav_in,
                              wav_out):
    x, sr = torchaudio.load(wav_in)
    x = x.detach().numpy()
    out = distort_wav_conf(x, distort_type, distort_conf, rate)
    torchaudio.save(wav_out, torch.from_numpy(out), sr)


if __name__ == "__main__":
    distort_type = sys.argv[1]
    wav_in = sys.argv[2]
    wav_out = sys.argv[3]
    conf = None
    rate = 0.1
    if distort_type == 'new_jag_distortion':
        conf = {'mask_number': 4}
    elif distort_type == 'new_fence_distortion':
        conf = {'mask_number': 1, 'max_db': -30}
    elif distort_type == 'poly_distortion':
        conf = {'a': 4, 'm': 2, "n": 2}
    distort_wav_conf_and_save(distort_type, conf, rate, wav_in, wav_out)



# Copyright      2021  Xiaomi Corp.        (authors: Fangjun Kuang)
#
# See ../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pathlib import Path
from typing import List, Union

import k2
import sentencepiece as spm
import torch


class BpeCtcTrainingGraphCompiler(object):
    def __init__(
        self,
        lang_dir: Path,
        device: Union[str, torch.device] = "cpu",
        sos_token: str = "<sos/eos>",
        eos_token: str = "<sos/eos>",
    ) -> None:
        """
        Args:
          lang_dir:
            This directory is expected to contain the following files:

                - bpe.model
                - words.txt
          device:
            It indicates CPU or CUDA.
          sos_token:
            The word piece that represents sos.
          eos_token:
            The word piece that represents eos.
        """
        lang_dir = Path(lang_dir)
        model_file = lang_dir / "bpe.model"
        sp = spm.SentencePieceProcessor()
        sp.load(str(model_file))
        self.sp = sp
        self.word_table = k2.SymbolTable.from_file(lang_dir / "words.txt")
        self.device = device

        self.sos_id = self.sp.piece_to_id(sos_token)
        self.eos_id = self.sp.piece_to_id(eos_token)

        assert self.sos_id != self.sp.unk_id()
        assert self.eos_id != self.sp.unk_id()

    def texts_to_ids(self, texts: List[str]) -> List[List[int]]:
        """Convert a list of texts to a list-of-list of piece IDs.

        Args:
          texts:
            It is a list of strings. Each string consists of space(s)
            separated words. An example containing two strings is given below:

                ['HELLO ICEFALL', 'HELLO k2']
        Returns:
          Return a list-of-list of piece IDs.
        """
        return self.sp.encode(texts, out_type=int)

    def compile(
        self,
        piece_ids: List[List[int]],
        modified: bool = False,
    ) -> k2.Fsa:
        """Build a ctc graph from a list-of-list piece IDs.

        Args:
          piece_ids:
            It is a list-of-list integer IDs.
          modified:
           See :func:`k2.ctc_graph` for its meaning.
        Return:
          Return an FsaVec, which is the result of composing a
          CTC topology with linear FSAs constructed from the given
          piece IDs.
        """
        graph = k2.ctc_graph(piece_ids, modified=modified, device=self.device)
        return graph



# Copyright    2022  Xiaomi Corp.        (authors: Wei Kang)
#
# See ../../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from typing import List

import k2
import torch
import torch.nn as nn
from beam_search import Hypothesis, HypothesisList, get_hyps_shape
from decode_stream import DecodeStream

from icefall.decode import one_best_decoding
from icefall.utils import get_texts


def greedy_search(
    model: nn.Module,
    encoder_out: torch.Tensor,
    streams: List[DecodeStream],
    blank_penalty: float = 0.0,
) -> None:
    """Greedy search in batch mode. It hardcodes --max-sym-per-frame=1.

    Args:
      model:
        The transducer model.
      encoder_out:
        Output from the encoder. Its shape is (N, T, C), where N >= 1.
      streams:
        A list of Stream objects.
    """
    assert len(streams) == encoder_out.size(0)
    assert encoder_out.ndim == 3

    blank_id = model.decoder.blank_id
    context_size = model.decoder.context_size
    device = model.device
    T = encoder_out.size(1)

    decoder_input = torch.tensor(
        [stream.hyp[-context_size:] for stream in streams],
        device=device,
        dtype=torch.int64,
    )
    # decoder_out is of shape (N, 1, decoder_out_dim)
    decoder_out = model.decoder(decoder_input, need_pad=False)
    decoder_out = model.joiner.decoder_proj(decoder_out)

    for t in range(T):
        # current_encoder_out's shape: (batch_size, 1, encoder_out_dim)
        current_encoder_out = encoder_out[:, t : t + 1, :]  # noqa

        logits = model.joiner(
            current_encoder_out.unsqueeze(2),
            decoder_out.unsqueeze(1),
            project_input=False,
        )
        # logits'shape (batch_size,  vocab_size)
        logits = logits.squeeze(1).squeeze(1)

        if blank_penalty != 0.0:
            logits[:, 0] -= blank_penalty

        assert logits.ndim == 2, logits.shape
        y = logits.argmax(dim=1).tolist()
        emitted = False
        for i, v in enumerate(y):
            if v != blank_id:
                streams[i].hyp.append(v)
                emitted = True
        if emitted:
            # update decoder output
            decoder_input = torch.tensor(
                [stream.hyp[-context_size:] for stream in streams],
                device=device,
                dtype=torch.int64,
            )
            decoder_out = model.decoder(
                decoder_input,
                need_pad=False,
            )
            decoder_out = model.joiner.decoder_proj(decoder_out)


def modified_beam_search(
    model: nn.Module,
    encoder_out: torch.Tensor,
    streams: List[DecodeStream],
    num_active_paths: int = 4,
    blank_penalty: float = 0.0,
) -> None:
    """Beam search in batch mode with --max-sym-per-frame=1 being hardcoded.

    Args:
      model:
        The RNN-T model.
      encoder_out:
        A 3-D tensor of shape (N, T, encoder_out_dim) containing the output of
        the encoder model.
      streams:
        A list of stream objects.
      num_active_paths:
        Number of active paths during the beam search.
    """
    assert encoder_out.ndim == 3, encoder_out.shape
    assert len(streams) == encoder_out.size(0)

    blank_id = model.decoder.blank_id
    context_size = model.decoder.context_size
    device = next(model.parameters()).device
    batch_size = len(streams)
    T = encoder_out.size(1)

    B = [stream.hyps for stream in streams]

    for t in range(T):
        current_encoder_out = encoder_out[:, t].unsqueeze(1).unsqueeze(1)
        # current_encoder_out's shape: (batch_size, 1, 1, encoder_out_dim)

        hyps_shape = get_hyps_shape(B).to(device)

        A = [list(b) for b in B]
        B = [HypothesisList() for _ in range(batch_size)]

        ys_log_probs = torch.stack(
            [hyp.log_prob.reshape(1) for hyps in A for hyp in hyps], dim=0
        )  # (num_hyps, 1)

        decoder_input = torch.tensor(
            [hyp.ys[-context_size:] for hyps in A for hyp in hyps],
            device=device,
            dtype=torch.int64,
        )  # (num_hyps, context_size)

        decoder_out = model.decoder(decoder_input, need_pad=False).unsqueeze(1)
        decoder_out = model.joiner.decoder_proj(decoder_out)
        # decoder_out is of shape (num_hyps, 1, 1, decoder_output_dim)

        # Note: For torch 1.7.1 and below, it requires a torch.int64 tensor
        # as index, so we use `to(torch.int64)` below.
        current_encoder_out = torch.index_select(
            current_encoder_out,
            dim=0,
            index=hyps_shape.row_ids(1).to(torch.int64),
        )  # (num_hyps, encoder_out_dim)

        logits = model.joiner(current_encoder_out, decoder_out, project_input=False)
        # logits is of shape (num_hyps, 1, 1, vocab_size)

        logits = logits.squeeze(1).squeeze(1)

        if blank_penalty != 0.0:
            logits[:, 0] -= blank_penalty

        log_probs = logits.log_softmax(dim=-1)  # (num_hyps, vocab_size)

        log_probs.add_(ys_log_probs)

        vocab_size = log_probs.size(-1)

        log_probs = log_probs.reshape(-1)

        row_splits = hyps_shape.row_splits(1) * vocab_size
        log_probs_shape = k2.ragged.create_ragged_shape2(
            row_splits=row_splits, cached_tot_size=log_probs.numel()
        )
        ragged_log_probs = k2.RaggedTensor(shape=log_probs_shape, value=log_probs)

        for i in range(batch_size):
            topk_log_probs, topk_indexes = ragged_log_probs[i].topk(num_active_paths)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                topk_hyp_indexes = (topk_indexes // vocab_size).tolist()
                topk_token_indexes = (topk_indexes % vocab_size).tolist()

            for k in range(len(topk_hyp_indexes)):
                hyp_idx = topk_hyp_indexes[k]
                hyp = A[i][hyp_idx]

                new_ys = hyp.ys[:]
                new_token = topk_token_indexes[k]
                if new_token != blank_id:
                    new_ys.append(new_token)

                new_log_prob = topk_log_probs[k]
                new_hyp = Hypothesis(ys=new_ys, log_prob=new_log_prob)
                B[i].add(new_hyp)

    for i in range(batch_size):
        streams[i].hyps = B[i]


def fast_beam_search_one_best(
    model: nn.Module,
    encoder_out: torch.Tensor,
    processed_lens: torch.Tensor,
    streams: List[DecodeStream],
    beam: float,
    max_states: int,
    max_contexts: int,
    blank_penalty: float = 0.0,
) -> None:
    """It limits the maximum number of symbols per frame to 1.

    A lattice is first generated by Fsa-based beam search, then we get the
    recognition by applying shortest path on the lattice.

    Args:
      model:
        An instance of `Transducer`.
      encoder_out:
        A tensor of shape (N, T, C) from the encoder.
      processed_lens:
        A tensor of shape (N,) containing the number of processed frames
        in `encoder_out` before padding.
      streams:
        A list of stream objects.
      beam:
        Beam value, similar to the beam used in Kaldi..
      max_states:
        Max states per stream per frame.
      max_contexts:
        Max contexts pre stream per frame.
    """
    assert encoder_out.ndim == 3
    B, T, C = encoder_out.shape
    assert B == len(streams)

    context_size = model.decoder.context_size
    vocab_size = model.decoder.vocab_size

    config = k2.RnntDecodingConfig(
        vocab_size=vocab_size,
        decoder_history_len=context_size,
        beam=beam,
        max_contexts=max_contexts,
        max_states=max_states,
    )
    individual_streams = []
    for i in range(B):
        individual_streams.append(streams[i].rnnt_decoding_stream)
    decoding_streams = k2.RnntDecodingStreams(individual_streams, config)

    for t in range(T):
        # shape is a RaggedShape of shape (B, context)
        # contexts is a Tensor of shape (shape.NumElements(), context_size)
        shape, contexts = decoding_streams.get_contexts()
        # `nn.Embedding()` in torch below v1.7.1 supports only torch.int64
        contexts = contexts.to(torch.int64)
        # decoder_out is of shape (shape.NumElements(), 1, decoder_out_dim)
        decoder_out = model.decoder(contexts, need_pad=False)
        decoder_out = model.joiner.decoder_proj(decoder_out)
        # current_encoder_out is of shape
        # (shape.NumElements(), 1, joiner_dim)
        # fmt: off
        current_encoder_out = torch.index_select(
            encoder_out[:, t:t + 1, :], 0, shape.row_ids(1).to(torch.int64)
        )
        # fmt: on
        logits = model.joiner(
            current_encoder_out.unsqueeze(2),
            decoder_out.unsqueeze(1),
            project_input=False,
        )
        logits = logits.squeeze(1).squeeze(1)

        if blank_penalty != 0.0:
            logits[:, 0] -= blank_penalty

        log_probs = logits.log_softmax(dim=-1)
        decoding_streams.advance(log_probs)

    decoding_streams.terminate_and_flush_to_streams()

    lattice = decoding_streams.format_output(processed_lens.tolist())
    best_path = one_best_decoding(lattice)
    hyp_tokens = get_texts(best_path)

    for i in range(B):
        streams[i].hyp = hyp_tokens[i]



"""
@File  :do_compute_wer.py
@Author:Xuelong Geng
@Date  :2024/6/10 23:40
@Desc  :
"""
from gxl_ai_utils.utils import utils_file
true_path, hyp_path, output_dir = utils_file.do_get_commandline_param(3, ["true_path", "hyp_path", "output_dir"])
utils_file.do_compute_wer(true_text_path=true_path, hyp_text_path=hyp_path, output_dir=output_dir)



import sys
sys.path.insert(0,"../../../")

from gxl_ai_utils.utils import utils_file

root_output_dir = "/mnt/sfs/asr/test_data/test_sets_format_3000"
# for 3500chat
input_tar_path = "/mnt/sfs/asr/update_data/3500_chat/tmp"
output_raw_path =f"{root_output_dir}/3500_chat/data.list"
utils_file.do_convert_shards2raw(None, raw_data_list_path=output_raw_path, output_wav_dir_path=input_tar_path)
output_text_path = f'{root_output_dir}/3500_chat/text'
output_token_path =f"{root_output_dir}/3500_chat/speech_token"
dict_list = utils_file.load_dict_list_from_jsonl(output_raw_path)
text_dict = {}
token_dict = {}
for dict_i in dict_list:
    text_dict[dict_i['key']] = dict_i['txt']
    token_dict[dict_i['key']] = utils_file.do_convert_str_to_float_list(dict_i['speech_token'])
utils_file.write_dict_to_scp(text_dict, output_text_path)
utils_file.write_dict_to_scp(token_dict, output_token_path)

input_tar_path = "/mnt/sfs/asr/update_data/3500_asr/tmp"
output_raw_path =f"{root_output_dir}/3500_asr/data.list"
utils_file.do_convert_shards2raw(None, raw_data_list_path=output_raw_path, output_wav_dir_path=input_tar_path)
output_text_path = f'{root_output_dir}/3500_asr/text'
output_token_path =f"{root_output_dir}/3500_asr/speech_token"
dict_list = utils_file.load_dict_list_from_jsonl(output_raw_path)
text_dict = {}
token_dict = {}
for dict_i in dict_list:
    text_dict[dict_i['key']] = dict_i['txt']
    token_dict[dict_i['key']] = utils_file.do_convert_str_to_float_list(dict_i['speech_token'])
utils_file.write_dict_to_scp(text_dict, output_text_path)
utils_file.write_dict_to_scp(token_dict, output_token_path)





import random
import numpy as np
import pyroomacoustics as pyr
# import gpuRIR
import os
import torchaudio
import soundfile as sf
import tqdm


def MC_RIR_generate(clean):
    target_sr = 16000

    # limit_rir_point = int(limit_rir_time * target_sr)

    while True:
        room_x = round(random.uniform(5, 8), 1)
        room_y = round(random.uniform(3, 5), 1)
        room_z = round(random.uniform(3, 4), 1)
        room_dim = [room_x, room_y, room_z]

        rt60_tgt = round(random.uniform(0.2, 1.2), 2)
        try:
            e_absorption, max_order = pyr.inverse_sabine(rt60_tgt, room_dim) #0.1614901799818631 95
        except Exception:
            print("Not able to generate rt60: %f, room_x is %f, room_y is %f, room_z is %f. Regenerating..." % (
            rt60_tgt, room_x, room_y, room_z))
            continue
        else:
            break

    while True:
        room = pyr.ShoeBox(room_dim, absorption=e_absorption, fs=target_sr, max_order=max_order)
        mic_location_x = room_x / 2.0
        mic_location = np.array([mic_location_x, 1.0, 1.5])
        room.add_microphone(mic_location)

        target_location_x = round(random.uniform(1, room_x - 1), 2)
        target_location_y = round(random.uniform(1, room_y - 1), 2)
        target_location_z = 1.5
        target_location = [target_location_x, target_location_y, target_location_z]

        try:
            room.add_source(target_location, signal=clean)
        except Exception:
            print("Source should be inside the room. Regenerating...")
            continue
        else:
            break

    rir_target = room.compute_rir()
    room.image_source_model()
    room.simulate()
    target = room.mic_array.signals
    rt60 = room.measure_rt60()

    return target, rir_target, room, rt60


# def gpusimulate(self, signal, rir, room):
#     if rir.ndim < 3:
#         rir = np.expand_dims(rir, axis=0)
#     rir = rir.astype(np.float32)
#     signal = signal.astype(np.float32)
#     signal = gpuRIR.simulateTrajectory(signal, rir)
#     signal = signal.T
#     return signal


def generate_signal(target_path, rirtarget_path):
    target, sr = sf.read(target_path)
    new_target, rir_target, room, rt60 = MC_RIR_generate(target)
    # target_mic = gpusimulate(target, rir_target, room)
    sf.write(rirtarget_path, new_target.T, sr)


from gxl_ai_utils.utils import utils_file

def little_handle(wav_dict,output_dir):
    for key in tqdm.tqdm(wav_dict.keys(),desc="handle_wav_scp",total=len(wav_dict)):
        wav_path = wav_dict[key]
        if key.endswith('.wav'):
            new_wav_name = key
        else:
            new_wav_name = key + '.wav'
        new_wav_path = os.path.join(output_dir, new_wav_name)
        generate_signal(wav_path, new_wav_path)


def handle_wav_scp(wav_scp_path,output_dir):
    utils_file.makedir_sil(output_dir)
    wav_dict = utils_file.load_dict_from_scp(wav_scp_path)
    wav_dict_list = utils_file.do_split_dict(wav_dict,7)
    # runner = utils_file.GxlFixedThreadPool(7)
    # for wav_dict_i in wav_dict_list:
    #     runner.add_thread(little_handle,[wav_dict_i,output_dir])
    # runner.start()
    little_handle(wav_dict_list[0],output_dir)



if __name__ == "__main__":
    import pdb; pdb.set_trace()
    now = utils_file.do_get_now_time()
    input_scp_path= "/home/work_nfs8/xlgeng/data/scp_test/aishell/wav.scp"
    handle_wav_scp(input_scp_path, 'data/wav_output/aishell_test')
    utils_file.logging_print('耗时: %s' % (utils_file.do_get_now_time() - now))


import os
import sys
sys.path.insert(0, '../../../../../')
import re

from gxl_ai_utils.utils import utils_file

import jieba
import tqdm
from nltk.translate.bleu_score import SmoothingFunction, sentence_bleu

def get_tag_from_str(input_str):
    # 使用正则表达式提取所有 <> 中的内容
    matches = re.findall(r'<.*?>', input_str)
    # 输出结果
    if len(matches)==0:
        return '<--no_tag-->'
    return matches[0].upper()
def convert_to_nearest_multiple(numbers, step=0.3):
    # 使用numpy的四舍五入功能将每个数字转换为最近的step的倍数
    return [round(num / step) * step for num in numbers]

def get_full_tag_from_str(input_str, smooth_interval=0.1):
    matches = re.findall(r'<.*?>', input_str)
    # 输出结果
    # if len(matches) == 0:
    #     return '<--no_tag-->'
    num_list = []
    for match in matches:
        try:
            # 尝试将匹配到的字符串转换为浮动数
            num_list.append(float(match[1:-1]))  # 去掉字符串的首尾字符后转换为 float
        except ValueError:
            # 如果转换失败，则将 -1 添加到 num_list
            num_list.append(-1)
    num_list = convert_to_nearest_multiple(num_list, step=smooth_interval)
    res_str = " ".join([f"{num:.6f}" for num in num_list])
    return res_str
def do_convert_to_pure_tag(input_file, output_file):
    text_dict = utils_file.load_dict_from_scp(input_file)
    new_text_dict = {}
    num_blank_tag = 0
    for key , value in text_dict.items():
        res_value = get_full_tag_from_str(value, smooth_interval=0.2)
        new_text_dict[key] =res_value
        if len(res_value) == 0:
            num_blank_tag += 1
    utils_file.logging_warning('空白时间戳的数量为：', num_blank_tag)
    output_file_blank_num = output_file + '_blank_num'
    utils_file.write_list_to_file([f'空白帧数量为{num_blank_tag}'],output_file_blank_num)
    utils_file.write_dict_to_scp(new_text_dict, output_file)
    return num_blank_tag


def do_showing_confusion_matrix(labels, matrix,title='', output_fig_path: str=None):
    """
    可视化混淆矩阵的函数
    :param labels: 标签序列，类型为列表等可迭代对象
    :param matrix: 代表混淆矩阵的二维列表，元素为整数，形状应为(len(labels), len(labels))
    """
    import matplotlib.pyplot as plt
    import numpy as np
    # 设置中文字体为黑体，以支持中文显示（确保系统中已安装黑体字体）
    plt.rcParams['font.sans-serif'] = ['SimHei']
    # 解决负号显示问题（有时候会出现负号显示异常，这一步是为了保证显示正常）
    plt.rcParams['axes.unicode_minus'] = False
    num_classes = len(labels)
    fig, ax = plt.subplots()
    # 使用imshow来绘制热力图展示混淆矩阵
    im = ax.imshow(np.array(matrix), cmap=plt.cm.Blues)

    # 设置x轴和y轴的刻度以及对应的标签
    ax.set_xticks(np.arange(num_classes))
    ax.set_yticks(np.arange(num_classes))
    ax.set_xticklabels(labels)
    ax.set_yticklabels(labels)

    # 旋转x轴刻度标签，让其更美观显示
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")

    # 添加每个方格中的数值标签
    for i in range(num_classes):
        for j in range(num_classes):
            text = ax.text(j, i, matrix[i][j],
                           ha="center", va="center", color="black")

    ax.set_title(f"{title} Confusion Matrix")
    ax.set_xlabel('Predicted label')
    ax.set_ylabel('True label')
    fig.tight_layout()
    if output_fig_path is not None:
        plt.savefig(output_fig_path)
    else:
        plt.show()

def calculate_bleu(candidate_text: str, reference_texts: list) -> float:
    """
    计算中文对话任务的 BLEU 分数。

    参数:
        candidate_text (str): 生成的文本。
        reference_texts (list): 参考文本列表（可以有多个参考句子）。

    返回:
        float: BLEU 分数。
    """
    # 对生成文本和参考文本进行分词
    candidate_tokens = list(jieba.cut(candidate_text))
    reference_tokens = [list(jieba.cut(ref)) for ref in reference_texts]

    # 使用 nltk 的 sentence_bleu 计算 BLEU 分数
    smoothing = SmoothingFunction().method1  # 避免分数为零
    bleu_score = sentence_bleu(reference_tokens, candidate_tokens,
                               weights=(0.25, 0.25, 0.25, 0.25),  # 1-gram 到 4-gram 的权重
                               smoothing_function=smoothing)

    return bleu_score

def do_compute_acc(ref_file, hyp_file, output_dir, output_dir2):
    """
    计算正确率和错误率，并列出混淆矩阵
    :return:
    """
    utils_file.makedir_sil(output_dir)
    utils_file.makedir_sil(output_dir2)
    utils_file.do_compute_wer(ref_file, hyp_file, output_dir)
    utils_file.logging_info('字符错误率计算完毕')
    wer_all = utils_file.do_get_wer_from_wer_file4all(os.path.join(output_dir, 'wer'))
    utils_file.copy_file(ref_file, os.path.join(output_dir2, 'ref_text'), use_shell=True)
    utils_file.copy_file(hyp_file, os.path.join(output_dir2, 'hyp_text'), use_shell=True)
    utils_file.copy_file(os.path.join(output_dir,'wer'), os.path.join(output_dir2, 'wer'))


    utils_file.logging_info('开始计算正确率，acc')
    acc_path = os.path.join(output_dir, 'acc')
    hyp_dict = utils_file.load_dict_from_scp(hyp_file)
    ref_dict = utils_file.load_dict_from_scp(ref_file)
    tag_hyp_dict = {}
    for key, value in hyp_dict.items():
        tag_hyp_dict[key] = get_tag_from_str(value)
    tag_ref_dict = {}
    for key, value in ref_dict.items():
        tag_ref_dict[key] = get_tag_from_str(value)
    output_acc_f = open(acc_path, 'w', encoding='utf-8')
    same_num = 0
    all_tags = set()
    # utils_file.print_dict(utils_file.do_get_random_subdict(tag_ref_dict, 10))
    for tag in tag_ref_dict.values():
        all_tags.add(tag)
    # for tag in tag_hyp_dict.values():
    #     all_tags.add(tag)
    labels = sorted(list(all_tags))
    # style_convert_dict = {
    #     "<新闻科普>": "<xinwei_kepu>",
    #     "<恐怖故事>": "<kongbu_gushi>",
    #     "<童话故事>": "<tonghua_gushi>",
    #     "<客服>": "<kefu>",
    #     "<诗歌散文>": "<shige_sanwen>",
    #     "<有声书>": "<youshengshu>",
    #     "<日常口语>": "<richang_kouyu>",
    #     "<其他>": "<qita>",
    # }

    utils_file.logging_info(f'标签种类为：{labels}')
    # labels = [style_convert_dict.get(label, label) for label in labels]
    utils_file.logging_info(f'标签种类为：{labels}')
    num_classes = len(labels)
    matrix = [[0] * num_classes for _ in range(num_classes)]
    for key, hyp_tags in tag_hyp_dict.items():
        if key not in tag_ref_dict:
            continue
        ref_tags = tag_ref_dict[key]
        # 判断标签是否相同
        if ref_tags == hyp_tags:
            if_same = True
            same_num += 1
            index = labels.index(ref_tags)
            matrix[index][index] += 1
        else:
            if_same = False
            # 标签不同的情况，找到真实标签和预测标签对应的索引，在相应的位置加1
            ref_index = labels.index(ref_tags)
            if hyp_tags in labels:
                hyp_index = labels.index(hyp_tags)
                matrix[ref_index][hyp_index] += 1
        # 向文件写入数据
        output_acc_f.write(f"key: {key}\n")
        output_acc_f.write(f"ref_tag: {ref_tags}\n")
        output_acc_f.write(f"hyp_tag: {hyp_tags}\n")
        output_acc_f.write(f"if_same: {if_same}\n")
        output_acc_f.write("\n")  # 添加空行分隔不同的条目
    acc_num = same_num / len(tag_hyp_dict)
    output_acc_f.write(f'正确率为：{acc_num}')
    output_acc_f.flush()
    output_acc_f.close()
    utils_file.copy_file(acc_path, os.path.join(output_dir2, 'acc'), use_shell=True)
    figure_path = os.path.join(output_dir, 'confusion_matrix.png')
    do_showing_confusion_matrix(labels, matrix, output_fig_path=figure_path)
    utils_file.copy_file(figure_path, os.path.join(output_dir2, 'confusion_matrix.png'))
    return {
        'acc': acc_num,
        'wer': wer_all,
    }

def do_compute_align(ref_file, hyp_file, output_dir, output_dir2):
    utils_file.makedir_sil(output_dir)
    utils_file.makedir_sil(output_dir2)
    utils_file.logging_info(f'开始计算的wer')
    utils_file.do_compute_wer(ref_file, hyp_file, output_dir)
    utils_file.logging_info('字符错误率计算完毕')
    wer_all = utils_file.do_get_wer_from_wer_file4all(os.path.join(output_dir, 'wer'))
    utils_file.copy_file(ref_file, os.path.join(output_dir2, 'ref_text'), use_shell=True)
    utils_file.copy_file(hyp_file, os.path.join(output_dir2, 'hyp_text'), use_shell=True)
    utils_file.copy_file(os.path.join(output_dir,'wer'), os.path.join(output_dir2, 'wer'))


    utils_file.logging_info(f'开始计算纯标签的wer_smooth')
    wer_pure_tag_smooth_path = os.path.join(output_dir, 'wer_pure_tag_smooth')
    hyp_file_tmp = utils_file.do_get_fake_file()
    ref_file_tmp = utils_file.do_get_fake_file()
    blank_num_hyp =  do_convert_to_pure_tag(hyp_file, hyp_file_tmp)
    blank_num_ref = do_convert_to_pure_tag(ref_file, ref_file_tmp)
    output_dir_tmp = '/home/xlgeng/.cache/'
    utils_file.do_compute_wer(ref_file_tmp, hyp_file_tmp, output_dir_tmp)
    utils_file.copy_file(os.path.join(output_dir_tmp, 'wer'), wer_pure_tag_smooth_path)
    utils_file.copy_file(wer_pure_tag_smooth_path, os.path.join(output_dir2, 'wer_pure_tag_smooth'))
    wer_smooth = utils_file.do_get_wer_from_wer_file4all(os.path.join(output_dir, 'wer_pure_tag_smooth'))
    utils_file.logging_info('纯标签的wer_smooth计算完毕')
    return {
        'wer': wer_all,
        'wer_smooth': wer_smooth,
        'blank_num_hyp': blank_num_hyp,
    }

from collections import defaultdict
import re
import edit_distance


class AverageShiftCalculator():
    def __init__(self):
        print("Calculating average shift.")

    def __call__(self, refs, hyps):
        ts_list1 = self.read_timestamps(refs)
        ts_list2 = self.read_timestamps(hyps)
        # print(ts_list1)
        # print(ts_list2)
        res = self.as_cal(ts_list1, ts_list2)
        print("Average shift : {}.".format(str(res)[:8]))
        print("Following timestamp pair differs most: {}, detail:{}".format(self.max_shift, self.max_shift_uttid))
        return res

    def _intersection(self, list1, list2):
        set1 = set(list1)
        set2 = set(list2)
        if set1 == set2:
            print("Uttid same checked.")
            return set1
        itsc = list(set1 & set2)
        print("Uttid differs: file1 {}, file2 {}, lines same {}.".format(len(list1), len(list2), len(itsc)))
        return itsc

    def read_timestamps(self, body_list):
        ts_list = []
        pattern_error = 0
        for body in body_list:
            body = body.replace("<|startoftranscript|>", "").replace("<|transcribe|>", "")
            ts_pattern = r"<\d{1,2}\.\d+>"
            if "<|en|>" in body:
                body = body.replace("<|en|>", "")
                lan = "en"
            elif "<|zh|>" in body:
                body = body.replace("<|zh|>", "")
                lan = "zh"

            all_time_stamps = re.findall(ts_pattern, body)
            all_time_stamps = [float(t.replace("<", "").replace(">", "")) for t in all_time_stamps]
            all_word_list = [x for x in re.split(ts_pattern, body)][1:-1]
            all_word_list = [x.strip() for x in all_word_list if x.strip() != '']

            if len(all_time_stamps) != 2 * len(all_word_list):
                pattern_error += 1
                continue
            text = "\t".join(all_word_list)
            ts = [all_time_stamps[i:i + 2] for i in range(0, len(all_time_stamps) - 1, 2)]
            ts_list.append((text, ts))
            assert len(ts) == len(all_word_list), f"{body}"
        print(f"pattern_error_num: {pattern_error}")
        return ts_list

    def _shift(self, filtered_timestamp_list1, filtered_timestamp_list2):
        shift_time = 0
        for fts1, fts2 in zip(filtered_timestamp_list1, filtered_timestamp_list2):
            shift_time += abs(fts1[0] - fts2[0]) + abs(fts1[1] - fts2[1])
        num_tokens = len(filtered_timestamp_list1)
        return shift_time, num_tokens

    def as_cal(self, ts_list1, ts_list2):
        # calculate average shift between timestamp1 and timestamp2
        # when characters differ, use edit distance alignment
        # and calculate the error between the same characters
        assert len(ts_list1) == len(ts_list2), f"{len(ts_list1)}, {len(ts_list2)}"
        self._accumlated_shift = 0
        self._accumlated_tokens = 0
        self.max_shift = 0
        self.max_shift_uttid = None
        for uttid in range(len(ts_list1)):
            (t1, ts1) = ts_list1[uttid]
            (t2, ts2) = ts_list2[uttid]
            # print(t1)
            # print(ts1)
            # print(t2)
            # print(ts2)
            # input()
            _align, _align2, _align3 = [], [], []
            fts1, fts2 = [], []
            _t1, _t2 = [], []
            sm = edit_distance.SequenceMatcher(t1.split('\t'), t2.split('\t'))
            s = sm.get_opcodes()
            # print("sm1",s)
            for j in range(len(s)):
                if s[j][0] == "replace" or s[j][0] == "insert":
                    _align.append(0)
                if s[j][0] == "replace" or s[j][0] == "delete":
                    _align3.append(0)
                elif s[j][0] == "equal":
                    _align.append(1)
                    _align3.append(1)
                else:
                    continue
            # use s to index t2
            for a, ts, t in zip(_align, ts2, t2.split('\t')):
                if a:
                    fts2.append(ts)
                    _t2.append(t)
            sm2 = edit_distance.SequenceMatcher(t2.split('\t'), t1.split('\t'))
            s = sm2.get_opcodes()
            for j in range(len(s)):
                if s[j][0] == "replace" or s[j][0] == "insert":
                    _align2.append(0)
                elif s[j][0] == "equal":
                    _align2.append(1)
                else:
                    continue
            # use s2 tp index t1
            for a, ts, t in zip(_align3, ts1, t1.split('\t')):
                if a:
                    fts1.append(ts)
                    _t1.append(t)
            if len(fts1) == len(fts2):
                shift_time, num_tokens = self._shift(fts1, fts2)
                if num_tokens == 0:
                    # print(ts_list1[uttid], ts_list2[uttid])
                    continue
                self._accumlated_shift += shift_time
                self._accumlated_tokens += num_tokens
                if shift_time / num_tokens > self.max_shift:
                    self.max_shift = shift_time / num_tokens
                    self.max_shift_uttid = uttid
            else:
                print("length mismatch")
        return self._accumlated_shift / self._accumlated_tokens

def _do_format_for_acc(input_file, output_file):
    pattern = re.compile(r'<\d+\.\d+>|[^\s<>]+')
    with open(input_file, 'r', encoding='utf-8') as infile, open(output_file, 'w', encoding='utf-8') as outfile:
        for line in infile:
            parts = line.strip().split(maxsplit=1)  # 分离ID和文本部分
            # print(parts)
            if len(parts) > 1:
                id_part = parts[0]
                text_part = parts[1]
                tokens = pattern.findall(text_part)

                # 将文本部分按照每3个字符插入空格
                try:
                    spaced_text = ' '.join(
                        [f"{tokens[i]}{tokens[i + 1]}{tokens[i + 2]}" for i in range(0, len(tokens), 3)])
                    # spaced_text = ' '.join([text_part[i:i+3] for i in range(0, len(text_part), 3)])
                    # 写入新文件
                    outfile.write(f'{id_part} {spaced_text}\n')
                except Exception as e:
                    print(f"跳过处理失败的行: {line.strip()}\n异常信息: {e}")
    print("格式化完成，输出文件为:", output_file)

def do_compute_align2(ref_file, hyp_file, output_dir, output_dir2):
    utils_file.makedir_sil(output_dir)
    utils_file.makedir_sil(output_dir2)
    utils_file.logging_info(f'开始计算的wer')
    utils_file.do_compute_wer(ref_file, hyp_file, output_dir)
    utils_file.logging_info('字符错误率计算完毕')
    wer_all = utils_file.do_get_wer_from_wer_file4all(os.path.join(output_dir, 'wer'))
    utils_file.copy_file(ref_file, os.path.join(output_dir2, 'ref_text'), use_shell=True)
    utils_file.copy_file(hyp_file, os.path.join(output_dir2, 'hyp_text'), use_shell=True)
    utils_file.copy_file(os.path.join(output_dir,'wer'), os.path.join(output_dir2, 'wer'))


    utils_file.logging_info(f'开始计算纯标签的ass指标')
    hyp_file_tmp = utils_file.do_get_fake_file()
    blank_num_hyp =  do_convert_to_pure_tag(hyp_file, hyp_file_tmp)

    output_famat_file_hyp = utils_file.do_get_fake_file()
    print('hyp_file',hyp_file)
    _do_format_for_acc(hyp_file, output_famat_file_hyp)
    print('ref_file',ref_file)
    output_famat_file_ref = utils_file.do_get_fake_file()
    _do_format_for_acc(ref_file, output_famat_file_ref)

    datas = defaultdict(list)
    refs = []
    hyps = []
    print(f'output_famat_file_ref,{output_famat_file_ref}')
    ref_keys = set()
    with open(output_famat_file_ref, 'r') as ref_timestamp:
        for line in ref_timestamp:
            tmp = line.strip().split(maxsplit=1)
            if tmp[0] not in ref_keys:
                ref_keys.add(tmp[0])
                datas[tmp[0]].append(tmp[1])
            else:
                continue
    print(f'output_famat_file_hyp,{output_famat_file_hyp}')
    hyp_keys = set()
    with open(output_famat_file_hyp, 'r') as hyp_timestamp:
        for line in hyp_timestamp:
            tmp = line.strip().split(maxsplit=1)
            if tmp[0] not in hyp_keys:
                hyp_keys.add(tmp[0])
                datas[tmp[0]].append(tmp[1])
            else:
                continue

    print('len datas ',len(datas))
    for utt, tmp in datas.items():
        if len(tmp) == 2:
            refs.append(tmp[0])
            hyps.append(tmp[1])


    # Create an instance of AverageShiftCalculator
    calculator = AverageShiftCalculator()

    # Call the calculator with the sample input data
    print('refs lens:', len(refs))
    aas_score = calculator(refs, hyps)
    print("aas_score", aas_score, '\n')


    utils_file.logging_info('ass计算完毕')
    return {
        'wer': wer_all,
        'ass': aas_score,
        'blank_num_hyp': blank_num_hyp,
    }



def do_compute_bleu_for_chat(ref_file, hyp_file, output_dir, output_dir2):
    # 计算BLEU得分
    utils_file.makedir_sil(output_dir)
    utils_file.makedir_sil(output_dir2)
    utils_file.copy_file(ref_file, os.path.join(output_dir2, 'ref_text'), use_shell=True)
    utils_file.copy_file(hyp_file, os.path.join(output_dir2, 'hyp_text'), use_shell=True)
    ref_dict = utils_file.load_dict_from_scp(ref_file)
    hyp_dict = utils_file.load_dict_from_scp(hyp_file)
    res_dict = {}
    total_bleu_score = 0
    for key, value in tqdm.tqdm(hyp_dict.items(), total=len(hyp_dict)):
        if key in ref_dict:
            ref_str = ref_dict[key]
            hyp_str = value
            bleu_score = calculate_bleu(hyp_str, [ref_str])
            res_dict[key] = bleu_score
            total_bleu_score += bleu_score
        else:
            print(f"key {key} not in ref_dict")
            continue

    avg_bleu_score = total_bleu_score / len(res_dict)
    res_dict['avg_bleu_score'] = avg_bleu_score
    utils_file.write_dict_to_scp(res_dict, os.path.join(output_dir, 'bleu'))
    utils_file.write_dict_to_scp(res_dict, os.path.join(output_dir2, 'bleu'))
    return {
        'avg_bleu_score': avg_bleu_score,
    }

if __name__ == '__main__':
    # 示例用法
    labels = ['class1', 'class2', 'class3', 'class4', 'class5']
    matrix = [[10, 2, 1, 0, 1],
              [3, 12, 1, 1, 0],
              [0, 0, 8, 0, 0],
              [1, 1, 0, 9, 0],
              [0, 0, 0, 1, 10]]
    do_showing_confusion_matrix(labels, matrix,title='age', output_fig_path='./output.png')


import os.path

from gxl_ai_utils.utils import utils_file


def copy_file():
    source_dir = "/mnt/disk1/yhdai/data_zipformer/batch03"
    output_dir = "/mnt/tencent/mnt/disk1/yhdai/data_zipformer/batch03"
    file_names_list = [
        'ximalaya_shenghuo_2.tar.gz',
        "wjtian_xmly2T_xuanyi_925h.tar.gz",
        "wjtian_xmly2T_renwen_3038h.tar.gz",
        "wjtian_lishi_2500h.tar.gz",
        "ximalaya_lishi_10T-1.tar.gz",
        "ximalaya_lishi_10T-0.tar.gz",
        "ximalaya_redian_2T.tar.gz",
        "zhguo_xmly2T_child_1350h.tar.gz",
    ]
    res_list = []
    for file_name in file_names_list:
        utils_file.logging_print(f'handling {file_name}')
        source_file_path_i = os.path.join(source_dir, file_name)
        output_file_path_i = os.path.join(output_dir, file_name)
        command_str = f'my_cp {source_file_path_i} {output_file_path_i}'
        res_list.append(command_str)
    utils_file.write_list_to_file(res_list, './run.sh')
if __name__ == '__main__':
    copy_file()


import glob
import os
import random

from tqdm import tqdm

from gxl_ai_utils.thread.my_thread import GxlDynamicThreadPool

from gxl_ai_utils.utils import utils_file, utils_data

base_path = "/home/work_nfs8/xlgeng/new_workspace/fairseq-main/examples/hubert/data_list"


def prepare_data():
    """
    将数据集转换为fairseq需要的格式
    Returns:

    """
    scp_path_list = utils_file.load_list_file_clean(
        "/home/work_nfs8/xlgeng/new_workspace/fairseq-main/examples/hubert/data_list/scp_path.txt")
    utils_file.print_list(scp_path_list)
    res_list = []
    for scp_path in scp_path_list:
        """"""
        utils_file.logging_print('开始处理: scp_path: {}'.format(scp_path))
        if not os.path.exists(scp_path):
            utils_file.logging_print("warning: scp_path not exists: {}".format(scp_path))
            continue
        dataset_name = (scp_path.split("/")[-2]).lower()
        wav_dict = utils_file.load_dict_from_scp(scp_path)
        utils_file.logging_print(f"dataset_name: {dataset_name}, lens: {len(wav_dict)}")
        # runner.add_thread(little_fun, [wav_dict, res_dict, dataset_name])
        res_list.extend(list(wav_dict.values()))

    res_file_path = "./added_data_gxl.tsv"
    random.shuffle(res_list)

    utils_file.logging_print('开始计算每个音频的采样点')
    num_thread = 100
    res_dict = {}

    def little_fun(little_list, res_dict):
        """"""
        for item in tqdm(little_list, total=len(little_list)):
            res_dict[item] = utils_data.get_sample_count(item)[0]

    res_list_list = utils_file.do_split_list(res_list, num_thread)
    thread_runner = GxlDynamicThreadPool()
    for res_list_i in res_list_list:
        thread_runner.add_thread(little_fun, [res_list_i, res_dict])
    thread_runner.start()

    utils_file.logging_print('开始写入: res_file_path: {}'.format(res_file_path))
    with open(res_file_path, "w") as f:
        f.write("/\n")
        for key, value in res_dict.items():
            f.write("{}\t{}\n".format(key, value))


def get_tsv(input_list_path, output_tsv_path, thread_num=100):
    """"""
    if isinstance(input_list_path, str):
        wav_list = utils_file.load_list_file_clean(input_list_path)
    elif isinstance(input_list_path, list):
        wav_list = input_list_path
    utils_file.logging_print('接着开始计算每个音频的采样点')
    num_thread = thread_num
    utils_file.logging_print('使用线程数: {}'.format(num_thread))
    res_dict = {}

    def little_fun(little_list, res_dict):
        """"""
        for item in tqdm(little_list, total=len(little_list)):
            res_dict[item] = utils_data.get_sample_count(item)[0]

    res_list_list = utils_file.do_split_list(wav_list, num_thread)
    thread_runner = GxlDynamicThreadPool()
    for res_list_i in res_list_list:
        thread_runner.add_thread(little_fun, [res_list_i, res_dict])
    thread_runner.start()

    utils_file.logging_print('开始写入: output_tsv_path: {}'.format(output_tsv_path))
    with open(output_tsv_path, "w") as f:
        f.write("/\n")
        for key, value in res_dict.items():
            f.write("{}\t{}\n".format(key, value))


def prepare_data_2():
    """得到全量数据的tsv和ltr"""
    data_dir_list = utils_file.load_list_file_clean(
        "./scp_path.txt")
    utils_file.logging_print('共有如下数据目录')
    utils_file.print_list(data_dir_list)
    utils_file.logging_print("对于每个数据目录，我们判断他们text wav.scp是否存在")
    for data_dir_i in data_dir_list:
        if not os.path.exists(data_dir_i):
            utils_file.logging_print("error：不存在: {}".format(data_dir_i))
            return
        wav_scp_path = os.path.join(data_dir_i, "wav.scp")
        text_path = os.path.join(data_dir_i, "text")
        if not os.path.exists(wav_scp_path):
            utils_file.logging_print("error：不存在: {}".format(wav_scp_path))
            return
        if not os.path.exists(text_path):
            utils_file.logging_print("error：不存在: {}".format(text_path))
            return
        with open(wav_scp_path, 'r') as file:
            line_count = sum(1 for line in file)
            utils_file.logging_print(f'{data_dir_i}的wav.scp行数为: {line_count}')
        with open(text_path, 'r') as file:
            line_count = sum(1 for line in file)
            utils_file.logging_print(f'{data_dir_i}的text行数为: {line_count}')

    utils_file.logging_print('判断通过')
    # utils_file.logging_print('开始处理数据, 首先合并所有的text和wav.scp')
    # big_text_dict = {}
    utils_file.logging_print('首先合并所有wav path')
    res_list = []
    for data_dir_i in data_dir_list:
        """"""
        wav_scp_path_i = os.path.join(data_dir_i, "wav.scp")
        wav_dict_i = utils_file.load_dict_from_scp(wav_scp_path_i)
        res_list.extend(list(wav_dict_i.values()))
    # res_file_path = "./added_data_and_part_wenetspeech.tsv"
    random.shuffle(res_list)
    utils_file.logging_print('开始得到10个wav_list_file，放到是个不同的节点上同时运行')
    res_list_list = utils_file.do_split_list(res_list, 10)
    for i, res_list_i in enumerate(res_list_list):
        res_file_path_i = f"./added_data_and_part_wenetspeech_{i}.list"
        utils_file.write_list_to_file(res_list_i, res_file_path_i)


def do_make_ltr_from_tsv_and_wav_text_scp(wav_dict_path, text_dict_path, tsv_path, output_tsv_path, output_ltr_path):
    """"""
    wav_dict = dict()
    text_dict = dict()
    tsv_list = list()
    if isinstance(wav_dict_path, str):
        wav_dict = utils_file.load_dict_from_scp(wav_dict_path)
    elif isinstance(wav_dict_path, dict):
        wav_dict = wav_dict_path
    if isinstance(text_dict_path, str):
        text_dict = utils_file.load_dict_from_scp(text_dict_path)
    elif isinstance(text_dict_path, dict):
        text_dict = text_dict_path
    if isinstance(tsv_path, str):
        tsv_list = utils_file.load_list_file_clean(tsv_path)
    elif isinstance(tsv_path, list):
        tsv_list = tsv_path
    if tsv_list[0] == "/":
        tsv_list = tsv_list[1:]
    all_info_dict = {}
    inverse_wav_dict = {v: k for k, v in wav_dict.items()}
    for tsv_i in tqdm(tsv_list, total=len(tsv_list)):
        """"""
        wav_path, sample_num = tsv_i.split()
        wav_key = inverse_wav_dict.get(wav_path, None)
        if wav_key is None:
            utils_file.logging_print(f'warning: {wav_path} not in wav_dict')
            continue
        text = text_dict.get(wav_key, None)
        if text is None:
            utils_file.logging_print(f'warning: {wav_key} not in text_dict')
            continue
        text = text.strip().replace(" ", "")
        text = " ".join(text)
        all_info_dict[tsv_i] = text
    utils_file.logging_print(f'all_info_dict: {len(all_info_dict)}')
    tsv_list_new = list(all_info_dict.keys())
    ltr_list_new = [all_info_dict[tsv_i] for tsv_i in tsv_list_new]
    tsv_list_new.insert(0, "/")
    utils_file.write_list_to_file(tsv_list_new, output_tsv_path)
    utils_file.write_list_to_file(ltr_list_new, output_ltr_path)


def get_ltr_for_added():
    """"""
    utils_file.logging_print('开始得到ltr文件')
    utils_file.logging_print('首先合并所有的小tsv')
    tsv_i_path_list = glob.glob("./added_data_and_part_wenetspeech_*.tsv")
    utils_file.print_list(tsv_i_path_list)
    all_info_list = []
    for tsv_i_path in tsv_i_path_list:
        """"""
        utils_file.logging_print('开始处理: tsv_i_path: {}'.format(tsv_i_path))
        with open(tsv_i_path, "r") as f:
            lines = f.readlines()
            lines = [line.strip() for line in lines]
            if lines[0] == "/":
                lines = lines[1:]
            all_info_list.extend(lines)
    random.shuffle(all_info_list)
    all_info_list.insert(0, "/")
    utils_file.write_list_to_file(all_info_list, "./added_data_and_part_wenetspeech.tsv")
    utils_file.logging_print('得到合并的tsv完成')

    utils_file.logging_print('开始得到总的wav_dict和text_dict')
    data_dir_list = utils_file.load_list_file_clean(
        "./scp_path.txt")
    wav_dict = {}
    text_dict = {}
    for data_dir_i in data_dir_list:
        if not os.path.exists(data_dir_i):
            utils_file.logging_print("error：不存在: {}".format(data_dir_i))
            return
        wav_scp_path = os.path.join(data_dir_i, "wav.scp")
        text_path = os.path.join(data_dir_i, "text")
        if not os.path.exists(wav_scp_path):
            utils_file.logging_print("error：不存在: {}".format(wav_scp_path))
            return
        if not os.path.exists(text_path):
            utils_file.logging_print("error：不存在: {}".format(text_path))
            return
        with open(wav_scp_path, 'r') as file:
            first_line = file.readline()
            first_wav_path = first_line.split()[1]
            if not os.path.exists(first_wav_path):
                utils_file.logging_print("error：音频位置不存在: {}".format(first_wav_path))
                return
        with open(wav_scp_path, 'r') as file:
            line_count = sum(1 for line in file)
            utils_file.logging_print(f'{data_dir_i}的wav.scp行数为: {line_count}')
        with open(text_path, 'r') as file:
            line_count = sum(1 for line in file)
            utils_file.logging_print(f'{data_dir_i}的text行数为: {line_count}')
        wav_dict_i = utils_file.load_dict_from_scp(wav_scp_path)
        text_dict_i = utils_file.load_dict_from_scp(text_path)
        wav_dict.update(wav_dict_i)
        text_dict.update(text_dict_i)

    utils_file.logging_print('判断通过')
    utils_file.logging_print(f"得到了wav_dict:{len(wav_dict)}, text_dict:{len(text_dict)}")
    do_make_ltr_from_tsv_and_wav_text_scp(wav_dict, text_dict, "./added_data_and_part_wenetspeech.tsv",
                                          "./added_data_and_part_wenetspeech_final.tsv",
                                          "./added_data_and_part_wenetspeech.ltr")

def get_ltr_for_trainl():
    """"""
    input_tsv = '/home/work_nfs9/lhli/dialect/data/wenetspeech/wenetspeech.tsv'
    wav_dict_path = "/home/work_nfs5_ssd/hfxue/data/data4w/source_1/train_l/wav.scp"
    text_dict_path = "/home/work_nfs5_ssd/hfxue/data/data4w/source_1/train_l/text"
    wav_dict = utils_file.load_dict_from_scp(wav_dict_path)
    text_dict = utils_file.load_dict_from_scp(text_dict_path)
    wav_dict = {k: v.replace('/home/work_nfs6/disk2/ASR_data/wav/train_l', '/home/38_data/ASR_data/wav/train_l') for
                k, v in tqdm(wav_dict.items(), total=len(wav_dict))}

    do_make_ltr_from_tsv_and_wav_text_scp(wav_dict, text_dict, input_tsv, "./trainl_final.tsv", "./trainl.ltr")


def add_2_tsv_and_ltr(i_tsv_path, i_ltr_path, j_tsv_path, j_ltr_path):
    i_tsv_list = utils_file.load_list_file_clean(i_tsv_path)
    if i_tsv_list[0] == "/":
        i_tsv_list = i_tsv_list[1:]
    i_ltr_list = utils_file.load_list_file_clean(i_ltr_path)

    j_tsv_list = utils_file.load_list_file_clean(j_tsv_path)
    if j_tsv_list[0] == "/":
        j_tsv_list = j_tsv_list[1:]

    j_ltr_list = utils_file.load_list_file_clean(j_ltr_path)
    tsv_list = i_tsv_list + j_tsv_list
    ltr_list = i_ltr_list + j_ltr_list
    if len(tsv_list) != len(ltr_list):
        utils_file.logging_print(f"error: len(tsv_list):{len(tsv_list)} != len(ltr_list):{len(ltr_list)}")
        return
    index_list = list(range(len(tsv_list)))
    random.shuffle(index_list)
    tsv_list = [tsv_list[i] for i in index_list]
    ltr_list = [ltr_list[i] for i in index_list]
    tsv_list.insert(0, "/")
    utils_file.write_list_to_file(tsv_list, "./final.tsv")
    utils_file.write_list_to_file(ltr_list, "./final.ltr")


def get_tsv_ltr_from_wav_text(wav_dict_path, text_dict_path, tsv_path, ltr_path):
    """"""
    wav_dict = utils_file.load_dict_from_scp(wav_dict_path)
    text_dict = utils_file.load_dict_from_scp(text_dict_path)
    wav_path_list = list(wav_dict.values())
    get_tsv(wav_path_list, tsv_path, 100)
    do_make_ltr_from_tsv_and_wav_text_scp(wav_dict, text_dict, tsv_path, tsv_path, ltr_path)



if __name__ == '__main__':
    # get_ltr_for_trainl()
    # add_2_tsv_and_ltr("./trainl_final.tsv",
    #                   "./trainl.ltr",
    #                   "/home/work_nfs8/xlgeng/new_workspace/gxl_ai_utils/eggs/cats_and_dogs/prepare_data_for_fairseq/added_data_and_part_wenetspeech_final.tsv",
    #                   "/home/work_nfs8/xlgeng/new_workspace/gxl_ai_utils/eggs/cats_and_dogs/prepare_data_for_fairseq/added_data_and_part_wenetspeech.ltr")
    # path_list = utils_file.load_list_file_clean('./final.tsv')
    # path_list.insert(0, '/')
    # utils_file.write_list_to_file(path_list, './final.tsv')
    get_tsv_ltr_from_wav_text("/home/backup_nfs2/nfs1_data/data_aishell/transcript/dev/wav.scp","/home/backup_nfs2/nfs1_data/data_aishell/transcript/dev/text", "./dev.tsv", "./dev.ltr")
    # i = 0
    # input_list_path = f"./added_data_and_part_wenetspeech_{i}.list"
    # input_list = utils_file.load_list_file_clean(input_list_path)
    # input_list_list = utils_file.do_split_list(input_list, 2)
    # for i, input_list_i in enumerate(input_list_list):
    #     utils_file.write_list_to_file(input_list_i, f"./added_data_and_part_wenetspeech_2_{i}.list")
    # input_list_path = f"./added_data_and_part_wenetspeech_2_{i}.list"
    # output_tsv_path = f"./added_data_and_part_wenetspeech_2_{i}.tsv"
    # get_tsv(input_list_path, output_tsv_path, thread_num=100)
